@@ -29,10 +29,11 @@ use LaTeXML::Core::MuGlue;
 use base qw(LaTeXML::Common::Object);
 #**********************************************************************
 sub new {
-  my ($class) = @_;
+  my ($class, %options) = @_;
   return bless {
-    mouth => undef, mouthstack => [], pushback => [], autoclose => 1, pending_comments => []
-    }, $class; }
+    mouth     => undef, mouthstack => [], pushback => [], autoclose => 1, pending_comments => [],
+    verbosity => $options{verbosity} || 0
+  }, $class; }
 
 #**********************************************************************
 # Start reading tokens from a new Mouth.
@@ -79,8 +80,8 @@ sub mouthIsOpen {
 # Corresponds (I think) to TeX's \endinput
 sub flushMouth {
   my ($self) = @_;
-  $$self{mouth}->finish;;    # but not close!
-  $$self{pushback}  = [];    # And don't read anytyhing more from it.
+  $$self{mouth}->finish;    # but not close!
+  $$self{pushback}  = [];   # And don't read anytyhing more from it.
   $$self{autoclose} = 1;
   return; }
 
@@ -117,7 +118,7 @@ sub readingFromMouth {
       $self->closeMouth(1); last; }
     elsif (!@{ $$self{mouthstack} }) {
       Error('unexpected', '<closed>', $self, "Mouth is unexpectedly already closed",
-        "Reading from " . Stringify($mouth) . ", but it has already been closed."); }
+        "Reading from " . Stringify($mouth) . ", but it has already been closed."); last; }
     elsif (!$$self{autoclose} || @{ $$self{pushback} } || $$self{mouth}->hasMoreInput) {
       my $next = Stringify($self->readToken);
       Error('unexpected', $next, $self, "Unexpected input remaining: '$next'",
@@ -130,48 +131,47 @@ sub readingFromMouth {
 
 # User feedback for where something (error?) occurred.
 sub getLocator {
-  my ($self, $long) = @_;
-  my $mouth = $$self{mouth};
-  my $i     = 0;
-  while ((defined $mouth) && (($$mouth{source} || '') eq 'Anonymous String')
+  my ($self) = @_;
+  my $mouth  = $$self{mouth};
+  my $i      = 0;
+  while ((defined $mouth) && (!defined $$mouth{source})
     && ($i < scalar(@{ $$self{mouthstack} }))) {
     $mouth = $$self{mouthstack}[$i++][0]; }
-  my $loc = (defined $mouth ? $mouth->getLocator($long) : '');
-  if (!$loc || $long) {
-    $loc .= show_pushback($$self{pushback}) if $long;
-    foreach my $frame (@{ $$self{mouthstack} }) {
-      my $ml = $$frame[0]->getLocator($long);
-      $loc .= ' ' . $ml if $ml;
-      last if $loc && !$long;
-      $loc .= show_pushback($$frame[1]) if $long; } }
-  return $loc; }
+  my $loc = (defined $mouth ? $mouth->getLocator : undef);
+  return $loc if defined $loc;
+  foreach my $frame (@{ $$self{mouthstack} }) {
+    my $ml = $$frame[0]->getLocator;
+    return $ml if defined $ml; }
+  return; }
 
 sub getSource {
   my ($self) = @_;
   my $source = defined $$self{mouth} && $$self{mouth}->getSource;
-  if (!$source) {
+  if (!defined($source)) {
     foreach my $frame (@{ $$self{mouthstack} }) {
       $source = $$frame[0]->getSource;
-      last if $source; } }
+      last if defined($source); } }
   return $source; }
 
 sub getSourceMouth {
   my ($self) = @_;
-  my $mouth = $$self{mouth};
+  my $mouth  = $$self{mouth};
   my $source = defined $mouth && $mouth->getSource;
-  if (!$source || ($source eq "Anonymous String")) {
+  if (!defined($source)) {
     foreach my $frame (@{ $$self{mouthstack} }) {
       $mouth  = $$frame[0];
       $source = $mouth->getSource;
-      last if $source && $source ne "Anonymous String"; } }
+      last if defined($source); } }
   return $mouth; }
 
 # Handy message generator when we didn't get something expected.
 sub showUnexpected {
   my ($self) = @_;
-  my $token = $self->readToken;
-  my $message = ($token ? "Next token is " . Stringify($token) : "Input is empty");
-  $self->unread($token);
+  my $message = "Input is empty";
+  if (my $token = $self->readToken) {
+    $message = "Next token is " . Stringify($token);
+    unshift(@{ $$self{pushback} }, $token);
+  }
   return $message; }
 
 sub show_pushback {
@@ -180,47 +180,38 @@ sub show_pushback {
   @pb = (@pb[0 .. 50], T_OTHER('...')) if scalar(@pb) > 55;
   return (@pb ? "\n  To be read again " . ToString(Tokens(@pb)) : ''); }
 
-#**********************************************************************
-# Not really 100% sure how this is supposed to work
-# See TeX Ch 20, p216 regarding noexpand, \edef with token list registers, etc.
-# Solution: Duplicate param tokens, stick NOTEXPANDED infront of expandable tokens.
-sub neutralizeTokens {
-  my ($self, @tokens) = @_;
-  my @result = ();
-  foreach my $token (@tokens) {
-    if ($$token[1] == CC_PARAM) {    # Inline ->getCatcode!
-      push(@result, $token); }
-    elsif (defined(my $defn = $STATE->lookupDefinition($token))) {
-      push(@result, Token('\noexpand', CC_NOTEXPANDED)); }
-    push(@result, $token); }
-  return @result; }
-
 #**********************************************************************
 # Low-level readers: read token, read expanded token
 #**********************************************************************
 # Note that every char (token) comes through here (maybe even twice, through args parsing),
 # So, be Fast & Clean!  This method only reads from the current input stream (Mouth).
+our @CATCODE_HOLD = (
+  0, 0, 0, 0,
+  0, 0, 0, 0,
+  0, 0, 0, 0,
+  0, 0, 1, 0,
+  0, 1, 0, 0);
+
 sub readToken {
   my ($self) = @_;
-  #  my $token = shift(@{$$self{pushback}});
   my $token;
   my $cc;
   # Check in pushback first....
-  while (defined($token = shift(@{ $$self{pushback} }))
-    && ($cc = $$token[1])
-    && (($cc == CC_COMMENT) || ($cc == CC_MARKER))) {    # NOTE: Inlined
+  while (($token = shift(@{ $$self{pushback} }))
+    && (($$token[1] != CC_SMUGGLE_THE) || ($token = $$token[2]))
+    && $CATCODE_HOLD[$cc = $$token[1]]) {
     if ($cc == CC_COMMENT) {
       push(@{ $$self{pending_comments} }, $token); }
     elsif ($cc == CC_MARKER) {
-      LaTeXML::Core::Definition::stopProfiling($token); } }
-  return $token if defined $token;
-  while (defined($token = $$self{mouth}->readToken())
-    && ($cc = $$token[1])
-    && (($cc == CC_COMMENT) || ($cc == CC_MARKER))) {    # NOTE: Inlined
+      LaTeXML::Core::Definition::stopProfiling($token, 'expand'); } }
+  if (defined $token) {
+    return $token; }
+  # Not in pushback, use the current mouth
+  while (($token = $$self{mouth}->readToken()) && $CATCODE_HOLD[$cc = $$token[1]]) {
     if ($cc == CC_COMMENT) {
-      push(@{ $$self{pending_comments} }, $token); }     # What to do with comments???
+      push(@{ $$self{pending_comments} }, $token); }    # What to do with comments???
     elsif ($cc == CC_MARKER) {
-      LaTeXML::Core::Definition::stopProfiling($token); } }
+      LaTeXML::Core::Definition::stopProfiling($token, 'expand'); } }
   return $token; }
 
 # Unread tokens are assumed to be not-yet expanded.
@@ -246,31 +237,44 @@ sub readXToken {
   return shift(@{ $$self{pending_comments} }) if $commentsok && @{ $$self{pending_comments} };
   my ($token, $cc, $defn);
   while (1) {
-    if (!defined($token = (@{ $$self{pushback} } ? shift(@{ $$self{pushback} }) : $$self{mouth}->readToken()))) {
+    $token = shift(@{ $$self{pushback} }) || $$self{mouth}->readToken();
+    if (!defined $token) {
       return unless $$self{autoclose} && $toplevel && @{ $$self{mouthstack} };
       $self->closeMouth; }    # Next input stream.
-    elsif (($cc = $$token[1]) == CC_NOTEXPANDED) {    # NOTE: Inlined ->getCatcode
-          # Should only occur IMMEDIATELY after expanding \noexpand (by readXToken),
-          # so this token should never leak out through an EXTERNAL call to readToken.
-      return $self->readToken; }    # Just return the next token.
-    elsif ($cc == CC_COMMENT) {
+    elsif (($cc = $$token[1]) == CC_SMUGGLE_THE) {
+      return $LaTeXML::SMUGGLE_THE ? $token : $$token[2]; }
+    elsif ($cc == CC_COMMENT) {    # NOTE: Inlined ->getCatcode
       return $token if $commentsok;
       push(@{ $$self{pending_comments} }, $token); }    # What to do with comments???
     elsif ($cc == CC_MARKER) {
-      LaTeXML::Core::Definition::stopProfiling($token); }
-    elsif (defined($defn = $STATE->lookupDefinition($token)) && $defn->isExpandable
-      && ($toplevel || !$defn->isProtected)) { # is this the right logic here? don't expand unless digesting?
+      LaTeXML::Core::Definition::stopProfiling($token, 'expand'); }
+    elsif (my $unexpanded = $$token[2]) {               # Inline get_dont_expand
+      return $token; }                                  # Defer expansion (recursion?)
+        # Note: special-purpose lookup in State, for efficiency
+    elsif (defined($defn = LaTeXML::Core::State::lookupExpandable($STATE, $token, $toplevel))) {
       local $LaTeXML::CURRENT_TOKEN = $token;
-      my $t;
-      # Do the check here, to be more forgiving and more informative
-      my @expansion = map { (($t = ref $_) eq 'LaTeXML::Core::Token' ? $_
-          : ($t eq 'LaTeXML::Core::Tokens' ? @$_
-            : (Error('misdefined', $token, undef,
-                "Expected a Token in expansion of " . ToString($token),
-                "got " . Stringify($_)), ()))) }
-        $defn->invoke($self);
-      # already checked tokens, so just push to be re-read (like ->unread(@expansion); )
+      my $invoked   = $defn->invoke($self) || [];
+      my @expansion = ();
+      for my $exp_t (@$invoked) {
+        my $r = ref $exp_t;
+        if ($r eq 'LaTeXML::Core::Token') {
+          push @expansion, $exp_t; }
+        elsif ($r eq 'LaTeXML::Core::Tokens') {
+          push @expansion, @$exp_t; }
+        else {
+          Fatal('misdefined', $r, undef, "Expected a Token, got " . Stringify($_)); } }
+      next unless @expansion;
+      if ($$LaTeXML::Core::Token::SMUGGLE_THE_COMMANDS{ $$defn{cs}[0] }) {
+        # magic THE_TOKS handling, add to pushback with a single-use noexpand flag only valid
+        #    at the exact time
+        # the token leaves the pushback.
+        # This is *required to be different* from the noexpand flag, as per the B Book
+        @expansion = map { T_SMUGGLE_THE($_); } @expansion; }
+      # add the newly expanded tokens back into the gullet stream, in the ordinary case.
       unshift(@{ $$self{pushback} }, @expansion); }
+    elsif ($cc == CC_CS && !(LaTeXML::Core::State::lookupMeaning($STATE, $token))) {
+      $STATE->generateErrorStub($self, $token);
+      return $token; }
     else {
       return $token; }    # just return it
   }
@@ -282,11 +286,11 @@ sub readRawLine {
   my ($self) = @_;
   # If we've got unread tokens, they presumably should come before the Mouth's raw data
   # but we'll convert them back to string.
-  my @tokens = @{ $$self{pushback} };
+  my @tokens  = map  { ($$_[1] == CC_SMUGGLE_THE ? $$_[2] : $_) } @{ $$self{pushback} };
   my @markers = grep { $_->getCatcode == CC_MARKER } @tokens;
   if (@markers) {    # Whoops, profiling markers!
-    @tokens = grep { $_->getCatcode != CC_MARKER } @tokens;            # Remove
-    map { LaTeXML::Core::Definition::stopProfiling($_) } @markers; }
+    @tokens = grep { $_->getCatcode != CC_MARKER } @tokens;                      # Remove
+    map { LaTeXML::Core::Definition::stopProfiling($_, 'expand') } @markers; }
   $$self{pushback} = [];
   # If we still have peeked tokens, we ONLY want to combine it with the remainder
   # of the current line from the Mouth (NOT reading a new line)
@@ -299,6 +303,9 @@ sub readRawLine {
 #**********************************************************************
 # Mid-level readers: checking and matching tokens, strings etc.
 #**********************************************************************
+# General note: TeX uses different tests for Space tokens in different places
+# (possibilities: catcode equality, ->equals, Equals and XEquals)
+
 # The following higher-level parsing methods are built upon readToken & unread.
 sub readNonSpace {
   my ($self) = @_;
@@ -307,16 +314,25 @@ sub readNonSpace {
   } while (defined $token && $$token[1] == CC_SPACE);    # Inline ->getCatcode!
   return $token; }
 
+sub readXNonSpace {
+  my ($self) = @_;
+  my $token;
+  do { $token = $self->readXToken(0);
+  } while (defined $token && $$token[1] == CC_SPACE);    # Inline ->getCatcode!
+  return $token; }
+
 sub skipSpaces {
   my ($self) = @_;
   my $tok = $self->readNonSpace;
-  $self->unread($tok) if defined $tok;
+  unshift(@{ $$self{pushback} }, $tok) if defined $tok;    # Unread
   return; }
 
+# Skip one space
+# if $expanded is true, it acts like <one optional space>, expanding the next token.
 sub skip1Space {
-  my ($self) = @_;
-  my $token = $self->readToken();
-  $self->unread($token) if $token && ($$token[1] != CC_SPACE);    # Inline ->getCatcode!
+  my ($self, $expanded) = @_;
+  my $token = ($expanded ? $self->readXToken : $self->readToken);
+  unshift(@{ $$self{pushback} }, $token) if $token && !Equals($token, T_SPACE);
   return; }
 
 # <filler> = <optional spaces> | <filler>\relax<optional spaces>
@@ -326,28 +342,55 @@ sub skipFiller {
     my $tok = $self->readNonSpace;
     return unless defined $tok;
     # Should \foo work too (where \let\foo\relax) ??
-    if ($tok->getString ne '\relax') { $self->unread($tok); return; }
+    if (!$tok->equals(T_CS('\relax'))) {
+      unshift(@{ $$self{pushback} }, $tok);    # Unread
+      return; }
   }
   return; }
 
 # Read a sequence of tokens balanced in {}
 # assuming the { has already been read.
 # Returns a Tokens list of the balanced sequence, omitting the closing }
+our @CATCODE_BALANCED_INTERESTING = (
+  0, 1, 1, 0,
+  0, 0, 0, 0,
+  0, 0, 0, 0,
+  0, 0, 0, 0,
+  0, 1, 0, 0);
+
 sub readBalanced {
-  my ($self) = @_;
+  my ($self, $expanded) = @_;
   my @tokens = ();
   my ($token, $level) = (undef, 1);
-  while ($level && defined($token = $self->readToken())) {
-    my $cc = $$token[1];    # Inline ->getCatcode!
-    $level++ if $cc == CC_BEGIN;
-    $level-- if $cc == CC_END;
-    push(@tokens, $token) if $level; }
-  return Tokens(@tokens); }
+  my $startloc = ($$self{verbosity} > 0) && $self->getLocator;
+  # Inlined readToken (we'll keep comments in the result)
+  while ($token = ($expanded ? $self->readXToken(0, 1) : $self->readToken())) {
+    my $cc = $$token[1];
+    if (!$CATCODE_BALANCED_INTERESTING[$cc]) {
+      push(@tokens, $token); }
+    elsif ($cc == CC_END) {
+      $level--;
+      if (!$level) {
+        last; }
+      push(@tokens, $token); }
+    elsif ($cc == CC_BEGIN) {
+      $level++;
+      push(@tokens, $token); }
+    elsif ($cc == CC_MARKER) {
+      LaTeXML::Core::Definition::stopProfiling($token, 'expand'); } }
+  if ($level > 0) {
+ # TODO: The current implementation has a limitation where if the balancing end is in a different mouth,
+ #       it will not be recognized.
+    my $loc_message = $startloc ? ("Started at " . ToString($startloc)) : ("Ended at " . ToString($self->getLocator));
+    Error('expected', "}", $self, "Gullet->readBalanced ran out of input in an unbalanced state.",
+      $loc_message); }
+  ## Performance enhancement:
+  return (wantarray ? (Tokens(@tokens), $token) : Tokens(@tokens)); }
 
 sub ifNext {
   my ($self, $token) = @_;
   if (my $tok = $self->readToken()) {
-    $self->unread($tok);
+    unshift(@{ $$self{pushback} }, $tok);    # Unread
     return $tok->equals($token); }
   else { return 0; } }
 
@@ -364,10 +407,10 @@ sub readMatch {
       if ($$token[1] == CC_SPACE) {    # If this was space, SKIP any following!!!
         while (defined($token = $self->readToken) && ($$token[1] == CC_SPACE)) {
           push(@matched, $token); }
-        $self->unread($token) if $token; }
+        unshift(@{ $$self{pushback} }, $token) if $token; }    # Unread
     }
-    return $choice unless @tomatch;    # All matched!!!
-    $self->unread(@matched);           # Put 'em back and try next!
+    return $choice unless @tomatch;                            # All matched!!!
+    unshift(@{ $$self{pushback} }, @matched);                  # Put 'em back and try next!
   }
   return; }
 
@@ -383,10 +426,10 @@ sub readKeyword {
     my @matched = ();
     my $tok;
     while (@tomatch && defined($tok = $self->readXToken(0)) && push(@matched, $tok)
-      && (uc($tok->getString) eq $tomatch[0])) {
+      && (uc($tok->toString) eq $tomatch[0])) {
       shift(@tomatch); }
-    return $keyword unless @tomatch;    # All matched!!!
-    $self->unread(@matched);            # Put 'em back and try next!
+    return $keyword unless @tomatch;             # All matched!!!
+    unshift(@{ $$self{pushback} }, @matched);    # Put 'em back and try next!
   }
   return; }
 
@@ -401,13 +444,37 @@ sub readUntil {
     push(@tokens, $token);
     $n++;
     if ($$token[1] == CC_BEGIN) {      # And if it's a BEGIN, copy till balanced END
-      push(@tokens, $self->readBalanced->unlist, T_END); } }
+      push(@tokens, $self->readBalanced); } }
   # Notice that IFF the arg looks like {balanced}, the outer braces are stripped
   # so that delimited arguments behave more similarly to simple, undelimited arguments.
   if (($n == 1) && ($tokens[0][1] == CC_BEGIN)) {
     shift(@tokens); pop(@tokens); }
   return (wantarray ? (Tokens(@tokens), $found) : Tokens(@tokens)); }
 
+sub readUntilBrace {
+  my ($self) = @_;
+  my @tokens = ();
+  my $token;
+  while (defined($token = $self->readToken())) {
+    if ($$token[1] == CC_BEGIN) {    # INLINE Catcode
+      unshift(@{ $$self{pushback} }, $token);    # Unread
+      last; }
+    push(@tokens, $token); }
+  return Tokens(@tokens); }
+
+# Skipping over conditional branches is used heavily when processing raw TeX (eg. tikz).
+# Make this efficient! Note that since we're skipping tokens, we presumably neither
+# need to note profiling end markers, nor preserve comments in the skipped sections?
+sub readNextConditional {
+  my ($self) = @_;
+  my $token;
+  my $type;
+  while ($token = shift(@{ $$self{pushback} }) || $$self{mouth}->readToken()) {
+    $token = $$token[2] if $$token[1] == CC_SMUGGLE_THE;
+    if ($type = $STATE->lookupConditional($token)) {
+      return ($token, $type); } }
+  return; }
+
 #**********************************************************************
 # Higher-level readers: Read various types of things from the input:
 #  tokens, non-expandable tokens, args, Numbers, ...
@@ -418,7 +485,7 @@ sub readArg {
   if (!defined $token) {
     return; }
   elsif ($$token[1] == CC_BEGIN) {    # Inline ->getCatcode!
-    return $self->readBalanced; }
+    return scalar($self->readBalanced(0)); }
   else {
     return Tokens($token); } }
 
@@ -432,7 +499,7 @@ sub readOptional {
   elsif (($tok->equals(T_OTHER('[')))) {
     return $self->readUntil(T_OTHER(']')); }
   else {
-    $self->unread($tok);
+    unshift(@{ $$self{pushback} }, $tok);    # Unread
     return $default; } }
 
 #**********************************************************************
@@ -458,11 +525,11 @@ sub readRegisterValue {
   my ($self, $type) = @_;
   my $token = $self->readXToken(0);
   return unless defined $token;
-  my $defn = $STATE->lookupDefinition($token);
+  my $defn = LaTeXML::Core::State::lookupDefinition($STATE, $token);
   if ((defined $defn) && ($defn->isRegister eq $type)) {
     return $defn->valueOf($defn->readArguments($self)); }
   else {
-    $self->unread($token);
+    unshift(@{ $$self{pushback} }, $token);    # Unread
     return; } }
 
 # Apparent behaviour of a token value (ie \toks#=<arg>)
@@ -471,21 +538,25 @@ sub readTokensValue {
   my $token = $self->readNonSpace;
   if (!defined $token) {
     return; }
-  elsif ($$token[1] == CC_BEGIN) {    # Inline ->getCatcode!
-    return $self->readBalanced; }
-  elsif (my $defn = $STATE->lookupDefinition($token)) {
+  elsif ($$token[1] == CC_BEGIN) {             # Inline ->getCatcode!
+    return scalar($self->readBalanced); }
+  elsif (my $defn = LaTeXML::Core::State::lookupDefinition($STATE, $token)) {
     if ($defn->isRegister eq 'Tokens') {
       return $defn->valueOf($defn->readArguments($self)); }
     elsif ($defn->isExpandable) {
-      $self->unread($defn->invoke($self));
+      if (my $x = $defn->invoke($self)) {
+        $self->unread(@{$x}); }
       return $self->readTokensValue; }
     else {
-      return $token; } }              # ?
+      return $token; } }                       # ?
   else {
     return $token; } }
 
 #======================================================================
 # some helpers...
+# Note that <one optional space> is kinda special:
+# The following Token(s) are expanded until an unexpandable token is found;
+# it is discarded if it is a space, but with an Equals() equality test!
 
 # <optional signs> = <optional spaces> | <optional signs><plus or minus><optional spaces>
 # return +1 or -1
@@ -493,18 +564,19 @@ sub readOptionalSigns {
   my ($self) = @_;
   my ($sign, $t) = ("+1", '');
   while (defined($t = $self->readXToken(0))
-    && (($t->getString eq '+') || ($t->getString eq '-') || ($t->equals(T_SPACE)))) {
+    && (($t->getString eq '+') || ($t->getString eq '-') || Equals($t, T_SPACE))) {
     $sign = -$sign if ($t->getString eq '-'); }
-  $self->unread($t) if $t;
+  unshift(@{ $$self{pushback} }, $t) if $t;    # Unread
   return $sign; }
 
+# Read digits (within $range), while expanding and if $skip, skip <one optional space> (expanded!)
 sub readDigits {
   my ($self, $range, $skip) = @_;
   my $string = '';
   my ($token, $digit);
-  while (($token = $self->readXToken(0)) && (($digit = $token->getString) =~ /^[$range]$/)) {
+  while (($token = $self->readXToken(0)) && (($digit = $token->toString) =~ /^[$range]$/)) {
     $string .= $digit; }
-  $self->unread($token) if $token && !($skip && $$token[1] == CC_SPACE);    # Inline ->getCatcode!
+  unshift(@{ $$self{pushback} }, $token) if $token && !($skip && Equals($token, T_SPACE));    #Inline
   return $string; }
 
 # <factor> = <normal integer> | <decimal constant>
@@ -518,10 +590,10 @@ sub readFactor {
     $string .= '.' . $self->readDigits('0-9');
     $token = $self->readXToken(0); }
   if (length($string) > 0) {
-    $self->unread($token) if $token && $$token[1] != CC_SPACE;    # Inline ->getCatcode!
+    unshift(@{ $$self{pushback} }, $token) if $token && $$token[1] != CC_SPACE; # Inline ->getCatcode, unread
     return $string; }
   else {
-    $self->unread($token);
+    unshift(@{ $$self{pushback} }, $token);                                     # Unread
     my $n = $self->readNormalInteger;
     return (defined $n ? $n->valueOf : undef); } }
 
@@ -535,15 +607,16 @@ sub readFactor {
 sub readNumber {
   my ($self) = @_;
   my $s = $self->readOptionalSigns;
-  if (defined(my $n = $self->readNormalInteger)) { return ($s < 0 ? $n->negate : $n); }
+  if    (defined(my $n = $self->readNormalInteger))  { return ($s < 0 ? $n->negate : $n); }
   elsif (defined($n = $self->readInternalDimension)) { return Number($s * $n->valueOf); }
   elsif (defined($n = $self->readInternalGlue))      { return Number($s * $n->valueOf); }
   else {
     my $next = $self->readToken();
-    $self->unread($next);
+    unshift(@{ $$self{pushback} }, $next);    # Unread
     Warn('expected', '<number>', $self, "Missing number, treated as zero",
       "while processing " . ToString($LaTeXML::CURRENT_TOKEN),
-      "next token is " . ToString($next));
+      "next token is " . ToString($next)
+      , " == " . Stringify($STATE->lookupMeaning($next)));
     return Number(0); } }
 
 # <normal integer> = <internal integer> | <integer constant>
@@ -552,21 +625,23 @@ sub readNumber {
 # Return a Number or undef
 sub readNormalInteger {
   my ($self) = @_;
-  my $token = $self->readXToken(0);
+  my $token = $self->readXToken(1);    # expand more
   if (!defined $token) {
     return; }
-  elsif (($$token[1] == CC_OTHER) && ($token->getString =~ /^[0-9]$/)) {    # Read decimal literal
+  elsif (($$token[1] == CC_OTHER) && ($token->toString =~ /^[0-9]$/)) {    # Read decimal literal
     return Number(int($token->getString . $self->readDigits('0-9', 1))); }
-  elsif ($token->equals(T_OTHER("\'"))) {                                   # Read Octal literal
+  elsif ($token->equals(T_OTHER("'"))) {                                   # Read Octal literal
     return Number(oct($self->readDigits('0-7', 1))); }
-  elsif ($token->equals(T_OTHER("\""))) {                                   # Read Hex literal
+  elsif ($token->equals(T_OTHER("\""))) {                                  # Read Hex literal
     return Number(hex($self->readDigits('0-9A-F', 1))); }
-  elsif ($token->equals(T_OTHER("\`"))) {                                   # Read Charcode
-    my $s = $self->readToken->getString;
+  elsif ($token->equals(T_OTHER("`"))) {                                   # Read Charcode
+    my $next = $self->readToken;
+    my $s    = ($next && $next->toString) || '';
     $s =~ s/^\\//;
+    $self->skip1Space(1);
     return Number(ord($s)); }    # Only a character token!!! NOT expanded!!!!
   else {
-    $self->unread($token);
+    unshift(@{ $$self{pushback} }, $token);    # Unread
     return $self->readInternalInteger; } }
 
 sub readInternalInteger {
@@ -587,10 +662,10 @@ sub readFloat {
     $token = $self->readXToken(0); }
   my $n;
   if (length($string) > 0) {
-    $self->unread($token) if $token && $$token[1] != CC_SPACE;    # Inline ->getCatcode!
+    unshift(@{ $$self{pushback} }, $token) if $token && $$token[1] != CC_SPACE; # Inline ->getCatcode, unread
     $n = $string; }
   else {
-    $self->unread($token) if $token;
+    unshift(@{ $$self{pushback} }, $token) if $token;                           # Unread
     $n = $self->readNormalInteger;
     $n = $n->valueOf if defined $n; }
   return (defined $n ? Float($s * $n) : undef); }
@@ -629,7 +704,7 @@ sub readDimension {
 sub readUnit {
   my ($self) = @_;
   if (defined(my $u = $self->readKeyword('ex', 'em'))) {
-    $self->skip1Space;
+    $self->skip1Space(1);
     return $STATE->convertUnit($u); }
   elsif (defined($u = $self->readInternalInteger)) {
     return $u->valueOf; }    # These are coerced to number=>sp
@@ -639,9 +714,10 @@ sub readUnit {
     return $u->valueOf; }
   else {
     $self->readKeyword('true');    # But ignore, we're not bothering with mag...
-    $u = $self->readKeyword('pt', 'pc', 'in', 'bp', 'cm', 'mm', 'dd', 'cc', 'sp');
+    my $units = $STATE->lookupValue('UNITS');
+    $u = $self->readKeyword(keys %$units);
     if ($u) {
-      $self->skip1Space;
+      $self->skip1Space(1);
       return $STATE->convertUnit($u); }
     else {
       return; } } }
@@ -677,7 +753,7 @@ sub readMuDimension {
 sub readMuUnit {
   my ($self) = @_;
   if (my $m = $self->readKeyword('mu')) {
-    $self->skip1Space;
+    $self->skip1Space(1);
     return $STATE->convertUnit($m); }
   elsif ($m = $self->readInternalMuGlue) {
     return $m->valueOf; }
@@ -775,7 +851,7 @@ sub readInternalMuGlue {
 
 __END__
 
-=pod 
+=pod
 
 =head1 NAME
 
@@ -814,7 +890,7 @@ Is this public? Clears all inputs.
 
 =item C<< $gullet->getLocator; >>
 
-Returns a string describing the current location in the input stream.
+Returns an object describing the current location in the input stream.
 
 =back
 
@@ -828,11 +904,6 @@ Return the L<LaTeXML::Core::Tokens> resulting from expanding all the tokens in C
 This is actually only used in a few circumstances where the arguments to
 an expandable need explicit expansion; usually expansion happens at the right time.
 
-=item C<< @tokens = $gullet->neutralizeTokens(@tokens); >>
-
-Another unusual method: Used for things like \edef and token registers, to
-inhibit further expansion of control sequences and proper spawning of register tokens.
-
 =item C<< $token = $gullet->readToken; >>
 
 Return the next token from the input source, or undef if there is no more input.
@@ -861,14 +932,16 @@ Read and return the next non-space token from the input after discarding any spa
 
 Skip the next spaces from the input.
 
-=item C<< $gullet->skip1Space; >>
+=item C<< $gullet->skip1Space($expanded); >>
 
 Skip the next token from the input if it is a space.
+If C($expanded> is true, expands ( like C< <one optional space> > ).
 
 =item C<< $tokens = $gullet->readBalanced; >>
 
 Read a sequence of tokens from the input until the balancing '}' (assuming the '{' has
-already been read). Returns a L<LaTeXML::Core::Tokens>.
+already been read). Returns a L<LaTeXML::Core::Tokens>,
+except in an array context, returns the collected tokens and the closing token.
 
 =item C<< $boole = $gullet->ifNext($token); >>
 
@@ -919,7 +992,7 @@ and return the value.  Returns undef if the next token isn't such a register.
 =item C<< $number = $gullet->readNumber; >>
 
 Read a L<LaTeXML::Common::Number> according to TeX's rules of the various things that
-can be used as a numerical value. 
+can be used as a numerical value.
 
 =item C<< $dimension = $gullet->readDimension; >>
 
@@ -953,4 +1026,3 @@ Public domain software, produced as part of work done by the
 United States Government & not subject to copyright in the US.
 
 =cut
-
