@@ -19,14 +19,19 @@ use charnames qw(:full);
 use LaTeXML::Post;
 use base qw(LaTeXML::Post::Processor);
 
+my $NBSP = pack('U', 0xA0);    # CONSTANT
+
 sub new {
   my ($class, %options) = @_;
   my $self = $class->SUPER::new(%options);
   $$self{db}       = $options{db};
   $$self{urlstyle} = $options{urlstyle};
 ##  $$self{toc_show} = ($options{number_sections} ? "typerefnum title" : "title");
-  $$self{toc_show}       = 'toctitle';
-  $$self{ref_show}       = ($options{number_sections} ? "typerefnum" : "title");
+  # Default format for ltx:ref's within TOC's
+  $$self{toc_show} = 'toctitle';
+  # Default format for regular ltx:ref's
+  # [BTW: Does number_sections really still make sense?]
+  $$self{ref_show}       = ($options{number_sections}        ? "refnum"                 : "title");
   $$self{min_ref_length} = (defined $options{min_ref_length} ? $options{min_ref_length} : 1);
   $$self{ref_join} = (defined $options{ref_join} ? $options{ref_join} : " \x{2023} "); # or " in " or ... ?
   $$self{navigation_toc} = $options{navigation_toc};
@@ -48,6 +53,9 @@ sub process {
   $self->fill_in_refs($doc);
   $self->fill_in_RDFa_refs($doc);
   $self->fill_in_bibrefs($doc);
+  $self->fill_in_mathlinks($doc);
+  $self->copy_resources($doc);
+
   if (keys %LaTeXML::Post::CrossRef::MISSING) {
     my $tempid = 0;
     foreach my $severity (qw(error warn info)) {
@@ -119,7 +127,7 @@ sub fill_in_relations {
           $doc->addNavigation($child_rel => $child_id); }
         else {                                # Else, consider it as some sort of sidebar.
           $doc->addNavigation('sidebar' => $child_id); } }
-    } }
+  } }
   return; }
 
 sub findPreviousPage {
@@ -136,7 +144,7 @@ sub findPreviousPage {
     # If there IS a preceding sibling, find it's rightmost descendant
     while (@sibs) {
       $pentry = $sibs[-1];
-      @sibs = grep { $_->getValue('primary') } $self->getChildPages($pentry); }
+      @sibs   = grep { $_->getValue('primary') } $self->getChildPages($pentry); }
     return $pentry; }                                                 # Return deepest page found
   return; }
 
@@ -166,6 +174,15 @@ sub getParentPage {
     && ($x = $x->getValue('pageid')) && ($x = $$self{db}->lookup("ID:" . $x))
     && $x; }
 
+sub getRootPage {
+  my ($self, $entry) = @_;
+  my $x    = $entry;
+  my $root = $entry;
+  while (($x = $x->getValue('parent')) && ($x = $$self{db}->lookup("ID:" . $x))
+    && ($x = $x->getValue('pageid')) && ($x = $$self{db}->lookup("ID:" . $x))) {
+    $root = $x; }
+  return $root; }
+
 # Assuming this entry is for a page, find the closest descendants that are (distinct) pages
 sub getChildPages {
   my ($self, $entry) = @_;
@@ -192,18 +209,29 @@ sub fill_in_tocs {
   foreach my $toc ($doc->findnodes('descendant::ltx:TOC[not(ltx:toclist)]')) {
     $n++;
     my $selector = $toc->getAttribute('select');
-    my $types    = ($selector
-      ? { map { ($_ => 1) } split(/\s*\|\s*/, $selector) }
-      : $normaltoctypes);
+    my $types;
+    if ($selector) {
+      $types = { map { ($_ => 1) } split(/\s*\|\s*/, $selector) }; }
     # global vs children of THIS or Document node?
     my $id     = $doc->getDocumentElement->getAttribute('xml:id');
-    my $format = $toc->getAttribute('format');
-
+    my $scope  = $toc->getAttribute('scope')  || 'current';
+    my $format = $toc->getAttribute('format') || 'normal';
+    my $lists;
+    if (my $listname = $toc->getAttribute('lists')) {
+      $lists = { map { $_ => 1 } split(/\s/, $listname) }; }
+    else {
+      $lists = { toc => 1 }; }
+    if ($scope eq 'global') {
+      if (my $entry = $$self{db}->lookup("ID:" . $id)) {
+        if (my $root = $self->getRootPage($entry)) {
+          $id = $root->getValue('pageid'); } } }
+    my $show = $toc->getAttribute('show') || $$self{toc_show};
     my @list = ();
-    if (!$format || ($format eq 'normal')) {
-      @list = $self->gentoc($id, $types); }
+    if (!$format || ($format =~ /^normal/)) {
+      @list = $self->gentoc($doc, $id, $show, $lists, $types); }
     elsif ($format eq 'context') {
-      @list = $self->gentoc_context($id, $types); }
+      $lists = { toc => 1 };
+      @list  = $self->gentoc_context($doc, $id, $show, $lists, $types); }
     $doc->addNodes($toc, ['ltx:toclist', {}, @list]) if @list; }
   NoteProgressDetailed(" [Filled in $n TOCs]");
   return; }
@@ -213,50 +241,67 @@ sub fill_in_tocs {
 # Returns a list of 0 or more ltx:tocentry's (possibly containing ltx:toclist's)
 # Note that parent/child relationships stored in ObjectDB can also reflect less
 # `interesting' objects like para or p style paragraphs, and such.
-#   $location: if defined (as a pathname), only include children that are on that page
-#   $depth   : only to the specific depth
-#
 sub gentoc {
-  my ($self, $id, $types, $localto, $selfid) = @_;
+  my ($self, $doc, $id, $show, $lists, $types, $localto, $selfid) = @_;
   if (my $entry = $$self{db}->lookup("ID:$id")) {
     my @kids = ();
     if ((!defined $localto) || (($entry->getValue('location') || '') eq $localto)) {
-      @kids = map { $self->gentoc($_, $types, $localto, $selfid) }
+      @kids = map { $self->gentoc($doc, $_, $show, $lists, $types, $localto, $selfid) }
         @{ $entry->getValue('children') || [] }; }
     my $type = $entry->getValue('type');
-    if ($$types{$type}) {
-      my $typename = $type; $typename =~ s/^ltx://;
-      return (['ltx:tocentry', (defined $selfid && ($selfid eq $id) ? { class => 'ltx_ref_self' } : {}),
-          ['ltx:ref', { show => 'toctitle', idref => $id }],
-          (@kids ? (['ltx:toclist', { class => "ltx_toc_$typename" }, @kids]) : ())]); }
+    my $role = $entry->getValue('role');
+    if (($types ? ($type = $entry->getValue('type')) && $$types{$type} : 1)
+      && inlist_match($lists, $entry->getValue('inlist'))) {
+      return $self->gentocentry($doc, $entry, $selfid, $show, @kids); }
     else {
       return @kids; } }
   else {
     return (); } }
 
+sub inlist_match {
+  my ($listsa, $listsb) = @_;
+  return ($listsa && $listsb && grep { $$listsb{$_} } keys %$listsa); }
+
+# Experimental show pattern:  before < filling > after
+sub gentocentry {
+  my ($self, $doc, $entry, $selfid, $show, @children) = @_;
+  my $id       = $entry->getValue('id');
+  my $type     = $entry->getValue('type');
+  my $typename = $type; $typename =~ s/^ltx://;
+  my ($before, $after);
+  if ($show =~ /^(.*?)\<(.*?)$/) { $before = $1; $show  = $2; }
+  if ($show =~ /^(.*?)\>(.*?)$/) { $show   = $1; $after = $2; }
+  # Good candidate for before = thumbnail
+  return (['ltx:tocentry',
+      { class => "ltx_tocentry_$typename"
+          . (defined $selfid && ($selfid eq $id) ? ' ltx_ref_self' : "") },
+      ($before ? $self->generateRef_simple($doc, $id, $before) : ()),
+      ['ltx:ref', { show => $show, idref => $id }],
+      ($after    ? $self->generateRef_simple($doc, $id, $after)                       : ()),
+      (@children ? (['ltx:toclist', { class => "ltx_toclist_$typename" }, @children]) : ())]); }
+
 # Generate a "context" TOC, that shows what's on the current page,
 # but also shows the page in the context of it's siblings & ancestors.
 # This is useful for putting in a navigation bar.
 sub gentoc_context {
-  my ($self, $id, $types) = @_;
+  my ($self, $doc, $id, $show, $lists, $types) = @_;
   if (my $entry = $$self{db}->lookup("ID:$id")) {
     # Generate Downward TOC covering items WITHIN the current page.
-    my @navtoc = $self->gentoc($id, $types, $entry->getValue('location') || '', $id);
+    my @navtoc = $self->gentoc($doc, $id, $show, $lists, $types, $entry->getValue('location') || '', $id);
     # Then enclose it upwards along with siblings & ancestors
     my $p_id;
     while (($p_id = $entry->getValue('parent')) && ($entry = $$self{db}->lookup("ID:$p_id"))) {
-      @navtoc = map { ($_ eq $id
+      @navtoc =
+        map {
+        ($_->getValue('id') eq $id
           ? @navtoc
-          : ['ltx:tocentry', {},
-            ['ltx:ref', { idref => $_, show => 'toctitle' }]]) }
-        grep { $$normaltoctypes{ $$self{db}->lookup("ID:$_")->getValue('type') } }
+          : $self->gentocentry($doc, $_, undef, $show)) }
+        grep { $$normaltoctypes{ $_->getValue('type') } }    # or should we use @inlist???
+        map  { $$self{db}->lookup("ID:$_") }
         @{ $entry->getValue('children') || [] };
-      my $type = $entry->getValue('type');
-      if ($$types{$type}) {
-        my $typename = $type; $typename =~ s/^ltx://;
-        @navtoc = (['ltx:tocentry', {},
-            ['ltx:ref', { show => 'toctitle', idref => $p_id }],
-            (@navtoc ? (['ltx:toclist', { class => "ltx_toc_$typename" }, @navtoc]) : ())]); }
+      if (($types ? $$types{ $entry->getValue('type') } : 1)
+        && $entry->getValue('parent')) {
+        @navtoc = ($self->gentocentry($doc, $entry, undef, $show, @navtoc)); }
       $id = $p_id; }
     return @navtoc; }
   else {
@@ -287,19 +332,20 @@ sub fill_in_refs {
     my $id   = $ref->getAttribute('idref');
     my $show = $ref->getAttribute('show');
     $show = $$self{ref_show} unless $show;
-    $show = $$self{toc_show} if ($show eq 'fulltitle') || ($show =~ /.+title|title.+/);
     if (!$id) {
       if (my $label = $ref->getAttribute('labelref')) {
         my $entry;
         if (($entry = $db->lookup($label)) && ($id = $entry->getValue('id'))) {
-          $show =~ s/^type//;       # Since author may have put explicit \S\ref... in!
-        }
+          $ref->setAttribute(idref => $id); }
         else {
           $self->note_missing('warn', 'Target for Label', $label);
+          my $cl = $ref->getAttribute('class');
+          $ref->setAttribute(class => ($cl ? $cl . ' ltx_missing_label' : 'ltx_missing_label'));
           if (!$ref->textContent) {
             $doc->addNodes($ref, $label);    # Just to reassure (?) readers.
             $ref->setAttribute(broken => 1); }
-        } } }
+    } } }
+
     if ($id) {
       $n++;
       if (!$ref->getAttribute('href')) {
@@ -313,7 +359,7 @@ sub fill_in_refs {
         $doc->addNodes($ref, $self->generateRef($doc, $id, $show)); }
       if (my $entry = $$self{db}->lookup("ID:$id")) {
         $ref->setAttribute(stub => 1) if $entry->getValue('stub'); }
-    } }
+  } }
   NoteProgressDetailed(" [Filled in $n refs]");
   return; }
 
@@ -329,10 +375,10 @@ sub fill_in_RDFa_refs {
         if (my $label = $ref->getAttribute($key . 'labelref')) {
           my $entry;
           if (($entry = $db->lookup($label)) && ($id = $entry->getValue('id'))) {
-          }
+            $ref->setAttribute($key . 'idref' => $id); }
           else {
             $self->note_missing('warn', "Target for $key Label", $label);
-          } } }
+      } } }
       if ($id) {
         $n++;
         if (!$ref->getAttribute($key)) {
@@ -341,11 +387,64 @@ sub fill_in_RDFa_refs {
               $ref->setAttribute($key => $url); } }
           else {
             $ref->setAttribute($key => '#' . $id); } }
-      } } }
+  } } }
   set_RDFa_prefixes($doc->getDocument, {});    # what prefixes??
   NoteProgressDetailed(" [Filled in $n RDFa refs]");
   return; }
 
+sub fill_in_mathlinks {
+  my ($self, $doc) = @_;
+  my $db = $$self{db};
+  my $n  = 0;
+  foreach my $sym ($doc->findnodes('descendant::*[@decl_id or @meaning]')) {
+    my $tag = $doc->getQName($sym);
+    next if $tag eq 'ltx:XMRef';               # Blech; list those TO fill-in, or list those to exclude?
+    next if $sym->hasAttribute('href');
+    my $decl_id = $sym->getAttribute('decl_id');
+    my $meaning = $sym->getAttribute('meaning');
+    my $entry;
+    if ($decl_id
+      && !$doc->findnodes('ancestor::ltx:glossaryphrase | ancestor::ltx:declare[@type]', $sym)) {
+      $entry = $$self{db}->lookup("DECLARATION:local:$decl_id"); }
+    elsif ($meaning) {
+      $entry = $$self{db}->lookup("DECLARATION:global:$meaning"); }
+    if ($entry) {
+      $n++;
+      ## HACK: DLMF copies $meaning to ltxx:meaning for search indexing
+      ## This should evolve into using (future) mml @mathrole?
+##      if ($meaning && $$doc{namespaces}{ltxx}) {
+##        $$node[1]{'ltxx:meaning'} = $meaning; }
+      if (my $id = $self->getIDForDeclaration($entry)) {    # Where defined
+        $sym->setAttribute(href => $self->generateURL($doc, $id));
+        if (my $tag = $entry->getValue('tag:short') || $entry->getValue('description')) {
+          $sym->setAttribute(title => getTextContent($doc, $tag)); }
+  } } }
+  NoteProgressDetailed(" [Filled in $n math links]");
+  return; }
+
+# Given a declaration entry (ltx:declare, or ltx:mark or ...)
+# Return the id of an appropriate link target.
+# Basically this is the parent, except (DLMF specific?) it should be a table ROW, not CELL
+# Or the numbered equationgroup, not the unnumbered equation
+sub getIDForDeclaration {
+  my ($self, $entry) = @_;
+  if (my $pid = $entry && $entry->getValue('parent')) {
+    if (my $pentry = $$self{db}->lookup("ID:$pid")) {
+      my $ptype = $pentry->getValue('type') || '';
+      # If definition is in a table cell, the correct id will be that of the row
+      if ($ptype eq 'ltx:td') {
+        if (my $gpid = $pentry->getValue('parent')) {
+          return $gpid; } }
+      # If definition is in unnumbered equation within an equation group, use id of the group
+      elsif (($ptype eq 'ltx:equation') && !$pentry->getValue('refnum')) {
+        if (my $gpid = $pentry->getValue('parent')) {
+          if (my $gpentry = $$self{db}->lookup("ID:$gpid")) {
+            my $gptype = $gpentry->getValue('type') || '';
+            if ($gptype eq 'ltx:equationgroup') {
+              return $gpid; } } } }
+    }
+    return $pid; } }
+
 # Needs to evolve into the combined stuff that we had in DLMF.
 # (eg. concise author/year combinations for multiple bibrefs)
 sub fill_in_bibrefs {
@@ -363,54 +462,69 @@ sub fill_in_bibrefs {
 sub make_bibcite {
   my ($self, $doc, $bibref) = @_;
 
-  # NOTE: bibkeys are downcased when we look them up!
-  my @keys         = map { lc($_) } grep { $_ } split(/,/, $bibref->getAttribute('bibrefs'));
+  my @keys         = grep { $_ } split(/,/, $bibref->getAttribute('bibrefs') || '');
   my $show         = $bibref->getAttribute('show');
   my @preformatted = $bibref->childNodes();
   if ($show && ($show eq 'none') && !@preformatted) {
     $show = 'refnum'; }
   if (!$show) {
-    map { $self->note_missing('info', "bibref 'show' parameter", $_) } @keys;
     $show = 'refnum'; }
   if ($show eq 'nothing') {    # Ad Hoc support for \nocite!t
     return (); }
-  my $sep   = $bibref->getAttribute('separator')   || ',';
-  my $yysep = $bibref->getAttribute('yyseparator') || ',';
-  my @phrases = $bibref->getChildNodes();    # get the ltx;bibrefphrase's in the bibref!
-                                             # Collect all the data from the bibliography
+  my $sep     = $bibref->getAttribute('separator')   || ',';
+  my $yysep   = $bibref->getAttribute('yyseparator') || ',';
+  my @phrases = element_nodes($bibref);    # get the ltx;bibrefphrase's in the bibref!
+                                           # Collect all the data from the bibliography
   my @data    = ();
+  my @lists   = split(/\s+/, $bibref->getAttribute('inlist') || 'bibliography');
   foreach my $key (@keys) {
-    if (my $bentry = $$self{db}->lookup("BIBLABEL:$key")) {
-      if (my $id = $bentry->getValue('id')) {
-        if (my $entry = $$self{db}->lookup("ID:$id")) {
-          my $authors  = $entry->getValue('authors');
-          my $fauthors = $entry->getValue('fullauthors');
-          my $keytag   = $entry->getValue('keytag');
-          my $year     = $entry->getValue('year');
-          my $typetag  = $entry->getValue('typetag');
-          my $number   = $entry->getValue('number');
-          my $title    = $entry->getValue('title');
-          my $refnum   = $entry->getValue('refnum');        # This come's from the \bibitem, w/o BibTeX
-          my ($rawyear, $suffix);
+    my ($bentry, $id, $entry);
+    # NOTE: bibkeys are downcased when we look them up!
+    foreach my $list (@lists) {            # Find the first of the lists that contains this bibkey
+      $bentry = $$self{db}->lookup("BIBLABEL:" . $list . ':' . lc($key));
+      last if $bentry; }
+    if ($bentry
+      && ($id    = $bentry->getValue('id'))
+      && ($entry = $$self{db}->lookup("ID:$id"))) {
+      my $authors  = $entry->getValue('authors');
+      my $fauthors = $entry->getValue('fullauthors');
+      my $keytag   = $entry->getValue('keytag');
+      my $year     = $entry->getValue('year');
+      my $typetag  = $entry->getValue('typetag');
+      my $number   = $entry->getValue('number');
+      my $title    = $entry->getValue('title');
+      my $refnum   = $entry->getValue('refnum');        # This come's from the \bibitem, w/o BibTeX
+      my ($rawyear, $suffix);
 
-          if ($year && ($year->textContent) =~ /^(\d\d\d\d)(\w)$/) {
-            ($rawyear, $suffix) = ($1, $2); }
-          $show = 'refnum' unless ($show eq 'none') || $authors || $fauthors || $keytag; # Disable author-year format!
-                                                                                         # fullnames ?
-          push(@data, { authors => [$doc->trimChildNodes($authors || $fauthors || $keytag)],
-              fullauthors => [$doc->trimChildNodes($fauthors || $authors || $keytag)],
-              authortext => ($authors || $fauthors ? ($authors || $fauthors)->textContent : ''),
-              year => [$doc->trimChildNodes($year || $typetag)],
-              rawyear => $rawyear,
-              suffix  => $suffix,
-              number  => [$doc->trimChildNodes($number)],
-              refnum  => [$doc->trimChildNodes($refnum)],
-              title   => [$doc->trimChildNodes($title || $keytag)],
-              attr    => { idref => $id,
-                href => orNull($self->generateURL($doc, $id)),
-                ($title ? (title => orNull($title->textContent)) : ()) } }); } } }
+      my $titlestring = undef;
+      if (defined $title) {
+        $titlestring = $title->textContent;
+        $titlestring =~ s/^\s+//;                       # Trim leading whitespace
+        $titlestring =~ s/\s+$//;                       # and trailing
+        $titlestring =~ s/\s+/ /gs; }                   # and normalize all other whitespace.
+      if ($year && ($year->textContent) =~ /^(\d\d\d\d)(\w)$/) {
+        ($rawyear, $suffix) = ($1, $2); }
+      $show = 'refnum' unless ($show eq 'none') || $authors || $fauthors || $keytag; # Disable author-year format!
+                                                                                     # fullnames ?
+      push(@data, {
+          key         => $key,
+          authors     => [$doc->trimChildNodes($authors  || $fauthors || $keytag)],
+          fullauthors => [$doc->trimChildNodes($fauthors || $authors  || $keytag)],
+          authortext  => ($authors || $fauthors ? ($authors || $fauthors)->textContent : ''),
+          year        => [$doc->trimChildNodes($year || $typetag)],
+          rawyear     => $rawyear,
+          suffix      => $suffix,
+          number      => [$doc->trimChildNodes($number)],
+          refnum      => [$doc->trimChildNodes($refnum)],
+          title       => [$doc->trimChildNodes($title || $keytag)],
+          attr        => { idref => $id,
+            href => orNull($self->generateURL($doc, $id)),
+            ($titlestring ? (title => $titlestring) : ()) } }); }
     else {
-      $self->note_missing('warn', 'Entry for citation', $key); } }
+      $self->note_missing('warn', 'Entry for citation', $key);
+      push(@data, { key => $key, refnum => [$key], title => [$key], year => [],
+          attr => { idref => $key, title => $key, class => "ltx_missing_citation" } });
+  } }
   my $checkdups = ($show =~ /author/i) && ($show =~ /(year|number)/i);
   my @refs      = ();
   my $saveshow  = $show;
@@ -421,47 +535,76 @@ sub make_bibcite {
     $show = $saveshow;
     if (($show eq 'none') && @preformatted) {
       @stuff = @preformatted; $show = ''; }
+    elsif ($$datum{attr}{class} && ($$datum{attr}{class} eq 'ltx_missing_citation')) {
+      @stuff  = (['ltx:ref', $$datum{attr}, $$datum{key}]);
+      $didref = 1;
+      $show   = '';
+    }
+
+    # Add delimeters for parsing...
+    $show =~ s/(\w)year/$1\{\}year/gi;
+    $show =~ s/(\w)phrase/$1\{\}phrase/gi;
     while ($show) {
-      if ($show =~ s/^authors?//i) {
-        push(@stuff, $doc->cloneNodes(@{ $$datum{authors} })); }
-      elsif ($show =~ s/^fullauthors?//i) {
-        push(@stuff, $doc->cloneNodes(@{ $$datum{fullauthors} })); }
-      elsif ($show =~ s/^title//i) {
-        push(@stuff, $doc->cloneNodes(@{ $$datum{title} })); }
-      elsif ($show =~ s/^refnum//i) {
-        push(@stuff, $doc->cloneNodes(@{ $$datum{refnum} })); }
-      elsif ($show =~ s/^phrase(\d)//i) {
-        push(@stuff, $phrases[$1 - 1]->childNodes) if $phrases[$1 - 1]; }
-      elsif ($show =~ s/^year//i) {
-        if (@{ $$datum{year} }) {
-          push(@stuff, ['ltx:ref', $$datum{attr}, @{ $$datum{year} }]);
+      if ($show =~ s/^(\w+)//) {
+        my $role = lc($1); $role =~ s/s$//;    # remove trailing plural
+        if ($role eq 'author') {
+          push(@stuff, $doc->cloneNodes(@{ $$datum{authors} })); }
+        elsif ($role eq 'fullauthor') {
+          push(@stuff, $doc->cloneNodes(@{ $$datum{fullauthors} })); }
+        elsif ($role eq 'title') {
+          push(@stuff, $doc->cloneNodes(@{ $$datum{title} })); }
+        elsif ($role eq 'refnum') {
+          push(@stuff, $doc->cloneNodes(@{ $$datum{refnum} })); }
+        elsif ($role =~ /^phrase(\d)$/) {
+          # HACK! Avoid empty () from situations where we've set the show (CITE_STYLE) too early
+          # and don't actually have author-year information!
+          my $n = $1;
+          if (($n == 1) && ($show =~ /^\s*year\s*phrase2/i) && !scalar(@{ $$datum{year} })
+            && (!$phrases[0] || (length($phrases[0]->textContent) <= 1))
+            && (!$phrases[1] || (length($phrases[1]->textContent) <= 1))) {
+            $show =~ s/^\s*year\s*phrase2//i; }
+          else {
+            push(@stuff, $phrases[$n - 1]->childNodes) if $phrases[$n - 1]; } }
+        elsif ($role eq 'year') {
+          if (!$$datum{year}) {
+            $self->note_missing('warn', 'Date for citation', $$datum{key}); }
+          elsif (@{ $$datum{year} }) {
+            push(@stuff, ['ltx:ref', $$datum{attr}, @{ $$datum{year} }]);
+            $didref = 1;
+            while ($checkdups && @data && ($$datum{authortext} eq $data[0]{authortext})) {
+              my $next = shift(@data);
+              push(@stuff, $yysep, ' ');
+              if ((($$datum{rawyear} || 'no_year_1') eq ($$next{rawyear} || 'no_year_2')) && $$next{suffix}) {
+                push(@stuff, ['ltx:ref', $$next{attr}, $$next{suffix}]); }
+              else {
+                push(@stuff, ['ltx:ref', $$next{attr}, @{ $$next{year} }]); } } } }
+        elsif ($role eq 'number') {
+          push(@stuff, ['ltx:ref', $$datum{attr}, @{ $$datum{number} }]);
           $didref = 1;
           while ($checkdups && @data && ($$datum{authortext} eq $data[0]{authortext})) {
             my $next = shift(@data);
-            push(@stuff, $yysep, ' ');
-            if ((($$datum{rawyear} || 'no_year_1') eq ($$next{rawyear} || 'no_year_2')) && $$next{suffix}) {
-              push(@stuff, ['ltx:ref', $$next{attr}, $$next{suffix}]); }
-            else {
-              push(@stuff, ['ltx:ref', $$next{attr}, @{ $$next{year} }]); } } } }
-      elsif ($show =~ s/^number//i) {
-        push(@stuff, ['ltx:ref', $$datum{attr}, @{ $$datum{number} }]);
-        $didref = 1;
-        while ($checkdups && @data && ($$datum{authortext} eq $data[0]{authortext})) {
-          my $next = shift(@data);
-          push(@stuff, $yysep, ' ', ['ltx:ref', $$next{attr}, @{ $$next{number} }]); } }
-      elsif ($show =~ s/^super//i) {
-        my @r = ();
-        push(@r, ['ltx:ref', $$datum{attr}, @{ $$datum{number} }]);
-        $didref = 1;
-        while ($checkdups && @data && ($$datum{authortext} eq $data[0]{authortext})) {
-          my $next = shift(@data);
-          push(@r, $yysep, ' ', ['ltx:ref', $$next{attr}, @{ $$next{number} }]); }
-        push(@stuff, ['ltx:sup', {}, @r]); }
-      elsif ($show =~ s/^(.)//) {
+            push(@stuff, $yysep, ' ', ['ltx:ref', $$next{attr}, @{ $$next{number} }]); } }
+        elsif ($role eq 'super') {
+          my @r = ();
+          push(@r, ['ltx:ref', $$datum{attr}, @{ $$datum{number} }]);
+          $didref = 1;
+          while ($checkdups && @data && ($$datum{authortext} eq $data[0]{authortext})) {
+            my $next = shift(@data);
+            push(@r, $yysep, ' ', ['ltx:ref', $$next{attr}, @{ $$next{number} }]); }
+          push(@stuff, ['ltx:sup', {}, @r]); }
+        else {
+          print STDERR "CITE ignoring show key '$role'\n"; } }
+      elsif ($show =~ s/^\{([^\}]*)\}//) {    # pass-thru literal, quoted with {}
+        push(@stuff, $1) if $1; }
+      elsif ($show =~ s/^~//) {               # Pass-thru spaces
+        push(@stuff, $NBSP) if @stuff; }
+      elsif ($show =~ s/^(\s+)//) {           # Pass-thru spaces
+        push(@stuff, $1) if @stuff; }
+      elsif ($show =~ s/^(\W+)//) {           # Pass-thru non show keywords
         push(@stuff, $1); } }
     push(@refs,
-      (@refs ? ($sep, ' ') : ()),
-      ($didref ? @stuff : (['ltx:ref', $$datum{attr}, @stuff]))); }
+      (@refs   ? ($sep, ' ') : ()),
+      ($didref ? @stuff      : (['ltx:ref', $$datum{attr}, @stuff]))); }
   return @refs; }
 
 sub generateURL {
@@ -476,10 +619,11 @@ sub generateURL {
       my $extension = $$self{extension} || 'xml';
       my $urlstyle  = $$self{urlstyle}  || 'file';
       if ($urlstyle eq 'server') {
-        $url =~ s/(^|\/)index.\Q$extension\E$/$1/; }    # Remove trailing index.$extension
+        # Remove trailing index.$extension but be careful not to leave url empty! (then it's "self")
+        $url =~ s/(^|\/)index.\Q$extension\E$/($1 ? $1 : '.\/')/e; }
       elsif ($urlstyle eq 'negotiated') {
-        $url =~ s/\.\Q$extension\E$//;                  # Remove trailing $extension
-        $url =~ s/(^|\/)index$/$1/;                     # AND trailing index
+        $url =~ s/\.\Q$extension\E$//;    # Remove trailing $extension
+        $url =~ s/(^|\/)index$/$1/;       # AND trailing index
       }
       $url = '.' unless $url;
       if (my $fragid = $object->getValue('fragid')) {
@@ -494,7 +638,6 @@ sub generateURL {
     $self->note_missing('warn', 'DB Entry for ID', $id); }
   return; }
 
-my $NBSP = pack('U', 0xA0);    # CONSTANT
 # Generate the contents of a <ltx:ref> of the given id.
 # show is a string containing substrings 'type', 'refnum' and 'title'
 # (standing for the type prefix, refnum and title of the id'd object)
@@ -511,19 +654,35 @@ sub generateRef {
       if (my @s = $self->generateRef_aux($doc, $entry, $show)) {
         push(@stuff, $pending) if $pending;
         push(@stuff, @s);
-        return @stuff if $self->checkRefContent(@stuff);
+        return @stuff if $self->checkRefContent($doc, @stuff);
         $pending = $$self{ref_join}; } # inside/outside this brace determines if text can START with the join.
       $id = $entry->getValue('parent'); } }
+  if (!@stuff) {                       # Try first child for a title-less document?
+    my $entry = $$self{db}->lookup("ID:$reqid");
+    if (($entry->getValue('type') || '') eq 'ltx:document') {
+      foreach my $c (@{ $entry->getValue('children') }) {
+        if (my $centry = $$self{db}->lookup("ID:$c")) {
+          if (my @s = $self->generateRef_aux($doc, $centry, $reqshow)) {
+            push(@stuff, @s); last; } } } } }
   if (@stuff) {
     return @stuff; }
   else {
     $self->note_missing('info', 'Usable title for ID', $reqid);
     return ($reqid); } }               # id is crummy, but better than "?"... or?
 
+# Just return the reqshow value for $reqid, or nothing
+sub generateRef_simple {
+  my ($self, $doc, $reqid, $reqshow) = @_;
+  my $pending = '';
+  my @stuff;
+  if (my $entry = $reqshow && $reqid && $$self{db}->lookup("ID:$reqid")) {
+    return $self->generateRef_aux($doc, $entry, $reqshow); }
+  return (); }
+
 # Check if the proposed content of a <ltx:ref> is "Good Enough"
 # (long enough, unique enough to give reader feedback,...)
 sub checkRefContent {
-  my ($self, @stuff) = @_;
+  my ($self, $doc, @stuff) = @_;
   # Length? having _some_ actual text ?
   my $s = text_content(@stuff);
   # Could compare a minum length
@@ -548,47 +707,61 @@ sub text_content_aux {
   else {
     return $n; } }
 
-# Interpret a "Show" pattern for a given DB entry.
-# The pattern can contain substrings to be substituted
-#   type   => the type prefix (eg Ch. or similar)
-#   refnum => the reference number
-#   title  => the title.
-# and any other random characters which are preserved.
+my %ref_fallbacks = (    # Alternative fields, when not found
+  typerefnum  => [qw(refnum)],
+  rrefnum     => [qw(typerefnum frefnum refnum)],    # obsolete?
+  toctitle    => [qw(title toccaption)],
+  title       => [qw(toccaption)],
+  rawtoctitle => [qw(toctitle title toccaption)],
+  rawtitle    => [qw(title toccaption)],
+);
+
+# Generate text to fill in an ltx:ref from a database entry for some object.
+# The show pattern indicates what data to use; usually a single keyword
+# (or keywords separated by spaces, ~ or {} enclosed literal text)
+# The keywords are things like refnum, title, caption, etc
+# (possibly coming from ltx:tag or other data; see Scan)
 sub generateRef_aux {
   my ($self, $doc, $entry, $show) = @_;
   my @stuff = ();
   my $OK    = 0;
-  $show =~ s/typerefnum\s*title/title/;    # Same thing NOW!!!
   while ($show) {
-    if ($show =~ s/^type(\.?\s*)refnum(\.?\s*)//) {
-      if (my $frefnum = $entry->getValue('frefnum') || $entry->getValue('refnum')) {
+    if ($show =~ s/^(\w+)//) {    # peel off next keyword
+      my $key   = lc($1);
+      my $class = ($key =~ /title/ ? 'ltx_ref_title' : 'ltx_ref_tag');
+      my @keys  = ($key, 'tag:' . $key,
+        ($ref_fallbacks{$key} ? @{ $ref_fallbacks{$key} } : ()));
+      my $value;
+      foreach my $k (@keys) {     # lookup the data for that keyword (or an alternative)
+        $value = $entry->getValue($k);
+        last if $value; }
+      if ($value) {
         $OK = 1;
-        push(@stuff, ['ltx:text', { class => 'ltx_ref_tag' }, $self->prepRefText($doc, $frefnum)]); } }
-    elsif ($show =~ s/^rrefnum(\.?\s*)//) {
-      if (my $refnum = $entry->getValue('rrefnum') || $entry->getValue('refnum')) {
-        $OK = 1;
-        push(@stuff, ['ltx:text', { class => 'ltx_ref_tag' }, $self->prepRefText($doc, $refnum)]); } }
-    elsif ($show =~ s/^refnum(\.?\s*)//) {
-      if (my $refnum = $entry->getValue('refnum')) {
-        $OK = 1;
-        push(@stuff, ['ltx:text', { class => 'ltx_ref_tag' }, $self->prepRefText($doc, $refnum)]); } }
-    elsif ($show =~ s/^toctitle//) {
-      if (my $title = $entry->getValue('toctitle') || $entry->getValue('title')
-        || $entry->getValue('toccaption')) {
-        $OK = 1;
-        push(@stuff, ['ltx:text', { class => 'ltx_ref_title' }, $self->prepRefText($doc, $title)]); } }
-    elsif ($show =~ s/^title//) {
-      if (my $title = $entry->getValue('title') || $entry->getValue('toccaption')) {    # !!!
-        $OK = 1;
-        push(@stuff, ['ltx:text', { class => 'ltx_ref_title' }, $self->prepRefText($doc, $title)]); } }
-    elsif ($show =~ s/^(.)//) {
+        push(@stuff, ['ltx:text', { class => $class }, $self->prepRefText($doc, $value)]); } }
+    elsif ($show =~ s/^\{([^\}]*)\}//) {    # pass-thru literal, quoted with {}
+      push(@stuff, $1) if $1; }
+    elsif ($show =~ s/^~//) {               # Pass-thru spaces
+      push(@stuff, $NBSP) if @stuff; }
+    elsif ($show =~ s/^(\s+)//) {           # Pass-thru spaces
+      push(@stuff, $1) if @stuff; }
+    elsif ($show =~ s/^(\W+)//) {           # Pass-thru non show keywords
       push(@stuff, $1); } }
+  # Maybe nothing found for this entry (probably retry on parent?)
   return ($OK ? @stuff : ()); }
 
 sub prepRefText {
   my ($self, $doc, $title) = @_;
   return $doc->cloneNodes($doc->trimChildNodes($self->fillInTitle($doc, $title))); }
 
+sub prepRawRefText {
+  my ($self, $doc, $title) = @_;
+  my $node = $self->prepRefText($doc, $title);
+  if ($doc->getQName($node) =~ /^ltx:(?:toc)title$/) {    # Trim tags from titles
+    my ($first) = element_nodes($node);
+    if ($first && ($doc->getQName($first) eq 'ltx:tag')) {
+      $node->removeChild($first); } }
+  return $node; }
+
 # Generate a title string for ltx:ref
 sub generateTitle {
   my ($self, $doc, $id) = @_;
@@ -597,12 +770,11 @@ sub generateTitle {
   my $altstring = "";
   while (my $entry = $id && $$self{db}->lookup("ID:$id")) {
     my $title = $self->fillInTitle($doc,
-      $entry->getValue('title') || $entry->getValue('rrefnum')
+###      $entry->getValue('title') || $entry->getValue('rrefnum')
+      $entry->getValue('title') || $entry->getValue('typerefnum')
         || $entry->getValue('frefnum') || $entry->getValue('refnum'));
     #    $title = $title->textContent if $title && ref $title;
     $title = getTextContent($doc, $title) if $title && ref $title;
-    $title =~ s/^\s+// if $title;
-    $title =~ s/\s+$// if $title;
     if ($title) {
       $string .= $$self{ref_join} if $string;
       $string .= $title; }
@@ -610,6 +782,14 @@ sub generateTitle {
   return $string || $altstring; }
 
 sub getTextContent {
+  my ($doc, $title) = @_;
+  $title = getTextContent_rec($doc, $title) if $title && ref $title;
+  $title =~ s/^\s+//s  if $title;    # Trim leading whitespace
+  $title =~ s/\s+$//s  if $title;    # and trailing
+  $title =~ s/\s+/ /gs if $title;    # and normalize all other whitespace.
+  return $title; }
+
+sub getTextContent_rec {
   my ($doc, $node) = @_;
   my $type = $node->nodeType;
   if ($type == XML_TEXT_NODE) {
@@ -618,12 +798,12 @@ sub getTextContent {
     my $tag = $doc->getQName($node);
     if ($tag eq 'ltx:tag') {
       return ($node->getAttribute('open') || '')
-        . $node->textContent    # assuming no nested ltx:tag
+        . $node->textContent         # assuming no nested ltx:tag
         . ($node->getAttribute('close') || ''); }
     else {
-      return join('', map { getTextContent($doc, $_); } $node->childNodes); } }
+      return join('', map { getTextContent_rec($doc, $_); } $node->childNodes); } }
   elsif ($type == XML_DOCUMENT_FRAG_NODE) {
-    return join('', map { getTextContent($doc, $_); } $node->childNodes); }
+    return join('', map { getTextContent_rec($doc, $_); } $node->childNodes); }
   else {
     return ''; } }
 
@@ -636,15 +816,13 @@ sub fillInTitle {
     next if $ref->textContent;
     my $show = $ref->getAttribute('show');
     $show = $$self{ref_show} unless $show;
-    $show = $$self{toc_show} if $show eq 'fulltitle';
     my $refentry;
     if (my $id = $ref->getAttribute('idref')) {
       $refentry = $$self{db}->lookup("ID:$id"); }
     elsif (my $label = $ref->getAttribute('labelref')) {
       $refentry = $$self{db}->lookup($label);
       if ($id = $refentry->getValue('id')) {
-        $refentry = $$self{db}->lookup("ID:$id"); }
-      $show =~ s/^type//; }    # Since author may have put explicit \S\ref... in!
+        $refentry = $$self{db}->lookup("ID:$id"); } }
     if ($refentry) {
       $doc->replaceNode($ref, $self->generateRef_aux($doc, $refentry, $show)); } }
   # Fill in (replace, actually) any embedded citations.
@@ -659,40 +837,74 @@ sub fillInGlossaryRef {
   my $n = 0;
   foreach my $ref ($doc->findnodes('descendant::ltx:glossaryref')) {
     $n++;
-    my $role = $ref->getAttribute('role') || '';
-    my $key  = $ref->getAttribute('key');
-    my $show = $ref->getAttribute('show');
-    if (my $entry = $$self{db}->lookup(join(':', 'GLOSSARY', $role, $key))) {
-      my $title = $entry->getValue('expansion');
+    my $key   = $ref->getAttribute('key');
+    my @lists = split(/\s+/, $ref->getAttribute('inlist') || 'glossary');
+    my $show  = $ref->getAttribute('show');
+    my ($list, $entry) = ('', undef);
+    foreach my $alist (@lists) {    # Find list with this key
+      if ($entry = $$self{db}->lookup(join(':', 'GLOSSARY', $alist, $key))) {
+        $list = $alist; last; } }
+    if ($entry) {
+      my $title = $entry->getValue('phrase:definition');
       if (!$ref->getAttribute('title') && $title) {
         $ref->setAttribute(title => $title->textContent); }
+      if (my $id = $entry->getValue('id')) {
+        $ref->setAttribute(idref => $id); }
       if (!$ref->textContent && !element_nodes($ref)) {
-        $doc->addNodes($ref, $self->generateGlossaryRefTitle($doc, $entry, $show)); }
-    } }
+        my @stuff = $self->generateGlossaryRefTitle($doc, $entry, $show);
+        if (@stuff) {
+          $doc->addNodes($ref, @stuff); }
+        else {
+          $self->note_missing('warn', "Glossary ($list) contents ($show) for key", $key);
+          $doc->addNodes($ref, $key);
+          $doc->addClass($ref, 'ltx_missing'); } } }
+    else {
+      $self->note_missing('warn', "Glossary ($list) Entry for key", $key); }
+    if (!$ref->textContent && !element_nodes($ref)) {
+      $doc->addNodes($ref, $key);
+      $doc->addClass($ref, 'ltx_missing'); } }
   NoteProgressDetailed(" [Filled in $n glossaryrefs]");
   return; }
 
 sub generateGlossaryRefTitle {
   my ($self, $doc, $entry, $show) = @_;
-  my @stuff = ();
-  my $OK    = 0;
-  while ($show) {
-    if ($show =~ s/^short//) {
-      if (my $phrase = $entry->getValue('phrase')) {
-        $OK = 1;
-        push(@stuff, ['ltx:text', { class => 'ltx_glossary_short' },
-            $self->prepRefText($doc, $phrase)]); } }
-    elsif ($show =~ s/^long//) {
-      if (my $phrase = $entry->getValue('expansion')) {
-        $OK = 1;
-        push(@stuff, ['ltx:text', { class => 'ltx_glossary_long' },
-            $self->prepRefText($doc, $phrase)]); } }
-    elsif ($show =~ s/^(.)//) {
-      push(@stuff, $1); } }
-  return ($OK ? @stuff : ()); }
+  my $phrases = $entry->getValue('phrases');
+  my @stuff   = ();
+  if (my $phrase = $entry->getValue('phrase:' . $show)) {
+    push(@stuff, ['ltx:text', { class => 'ltx_glossary_' . $show },
+        $self->prepRefText($doc, $phrase)]); }
+  elsif ($show =~ /^(\w+)-plural$/) {
+    my $sh = $1;
+    if (my $phrase = $entry->getValue('phrase:' . $sh)) {
+      push(@stuff, ['ltx:text', { class => 'ltx_glossary_' . $show },
+          $self->prepRefText($doc, $phrase), 's']); } }
+  elsif ($show =~ /^(\w+)-indefinite$/) {
+    my $sh = $1;
+    if (my $phrase = $entry->getValue('phrase:' . $sh)) {
+      my $s   = $phrase->textContent;
+      my $art = ($s =~ /^[aeiou]/i ? 'an ' : 'a ');
+      push(@stuff, ['ltx:text', { class => 'ltx_glossary_' . $show },
+          $art, $self->prepRefText($doc, $phrase)]); } }
+  return @stuff; }
 
 sub orNull {
   return (grep { defined } @_) ? @_ : undef; }
+
+# Possibly this needs support from Scan, as well?
+# to manage resources, record in Manifest, something like that?
+sub copy_resources {
+  my ($self, $doc) = @_;
+  # Copy any "resources" linked from the document
+  my $paths = [$doc->getSearchPaths];
+  foreach my $n ($doc->findnodes('//ltx:ref[@href and not(@idref) and not(@labelref)]')) {
+    my $url = $n->getAttribute('href');
+    if ($url !~ /^(\w+:|\/)/) {    # relative path? (No explicit protocol, or absolute)
+      if (my $src = pathname_find($url, paths => $paths)) {    # AND if file exists there.
+        my $dst = $doc->checkDestination($url);
+        pathname_copy($src, $dst);
+  } } }
+  return; }
+
 # ================================================================================
 1;
 
